/*
    futebus - FUse TExt BUS - fuse based IPC bus for text lines
    Copyright (C) 2023  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://www.repo.hu/projects/futebus
    VCS: svn://svn.repo.hu/futebus/trunk
    Contact: see doc/contact.html
*/

#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>

#define _FILE_OFFSET_BITS 64
#define FUSE_USE_VERSION 28
#include <fuse.h>

#include "futebus.h"
#include "message.h"
#include "address.h"
#include "channel.h"
#include "client.h"

#include "dispatch.h"

#define trace(c, ...) \
do { \
	if (c == NULL) \
		logprintf_("(NULL) "); \
	else \
		logprintf_("(%d) ", (int)c->id); \
	logprintf_(__VA_ARGS__);\
	logprintf("\n"); \
} while(0)

/* Worst case scenario: a client writes msg_size_max-1 without \n, then
   in a subsequent write() writes the \n and the next message. The write
   buffer never needs to host more than the first message plus some safety
   margin to make sure there was no \n (and we can throw an error for message
   too long). When writing multiple messages, this will lead to short writes. */
#define wbuf_size_max (msg_size_max + 8)


/* remove the first (oldest) message from rlist of c */
static message_t *rlist_pop(client_t *c)
{
	message_t *msg;

	mutex_lock(&c->rlist_lock);
	msg = c->rlist.array[0];
	vtp0_remove(&c->rlist, 0, 1);
	mutex_unlock(&c->rlist_lock);

	return msg;
}

/* handle "rlist is full" on write(). Returns:
    - negative on error (-errno, should be returned)
    - 0 when eof should be returned
    - 1 on success (message can be inserted now) */
static int rlist_is_full(client_t *dst, int can_block, int nonblock_unicast_write)
{
	if (can_block) { /* block write() until there's room */
		if (nonblock_unicast_write)
			return -EAGAIN;

		blocking_wait:;
		client_waiter_inc(dst); /* so dst doesn't get freed */
		for(;;) {
			if (dst->dead) {
				client_waiter_dec(dst);
				return -ESRCH;
			}
			if (dst->rlist.used < rlist_size_max) {
				client_waiter_dec(dst);
				return 1;
			}
			if (fuse_interrupted()) {
				client_waiter_dec(dst);
				return -EINTR;
			}
			usleep(100000); /* can't notify() here: multiple write()s may be pending */
		}
	}
	else if (dst->master) {
		/* when sending to a master: blocking wait until there's a free slot in there */
		goto blocking_wait;
	}
	else { /* make room for the new message by removing the oldest (drop msg) */
		message_t *old = rlist_pop(dst);
		message_unref(old);
		return 1;
	}

	return -EIO; /* shouldn't be able to get here */
}

int insert_msg_in_rlist(client_t *dst, message_t *msg, int nonblock_unicast_write, int can_block, int force)
{
	trace(dst, "write: insert rlist");

	/* check if rlist is full */
	if (!force && (dst->rlist.used >= rlist_size_max)) {
		int res = rlist_is_full(dst, nonblock_unicast_write, can_block);
		if (res <= 0)
			return res;
	}

	trace(dst, " rlist lock");
	mutex_lock(&dst->rlist_lock);
	vtp0_append(&dst->rlist, msg);

	trace(dst, " msg refco inc");
	mutex_lock(&msg->refco_lock);
	msg->refco++;
	mutex_unlock(&msg->refco_lock);

	trace(dst, " new rlist len: %ld", (long)dst->rlist.used);
	if (dst->rlist.used == 1) {
		trace(dst, " notify send to: %ld", (long)dst->id);
		notify_send(&dst->rlist_notify); /* the client may be waiting in a read() */
		futebus_poll_readable(dst); /* let an outstanding poll() know about POLLIN */
	}

	mutex_unlock(&dst->rlist_lock);

	trace(dst, " (fin)");
	return 1;
}

/* Figure if sending a message to a given channel would have any client
   reading it (returns 1 if so). If there's no reader that can receive
   the message (return value 0), the message should be discarded. On error
   returns a negative number that can be returned for write() */
static int write_dispatch_chan_has_readers(channel_t *chan, client_t *from, address_t dst_addr)
{
	client_t *c;

	switch(chan->type) {
		case 'A':
			if (!dst_addr.valid) return -EFAULT; /* must have a valid address */
			if (dst_addr.bcast) goto bcast;

			unicast:;
			for(c = gdl_first(&chan->clients); c != NULL; c = gdl_next(&chan->clients, c))
				if (c->id == dst_addr.id)
					return 1;
			return 0;

		case 'S':
			if (from->master) { /* master->slave */
				if (!dst_addr.valid) return -EFAULT; /* must have a valid address */
				if (dst_addr.bcast) goto bcast;
				goto unicast;
			}
			if (chan->master == NULL) return -ESRCH; /* can't write if there's no master */
			return 1; /* slave->master: we have a master so it is okay */

		case 'U':
			if (from->master) goto bcast; /* master->slave */
			if (chan->master == NULL) return -ESRCH; /* can't write if there's no master */
			return 1; /* slave->master: we have a master so it is okay */

		case 'B':
			bcast:;
			/* plain broadcast */
			for(c = gdl_first(&chan->clients); c != NULL; c = gdl_next(&chan->clients, c))
				if ((c != from) && c->rd)
					return 1;
			break;
	}
	return 0;
}

/* Deliver message to readers */
static int write_dispatch_chan_msg(channel_t *chan, client_t *from, message_t *msg, address_t dst_addr, int nonblock)
{
	client_t *c;
	int res = 0, unicast_can_block = 0;

	switch(chan->type) {
		case 'A':
			if (!dst_addr.valid) return -EFAULT; /* must have a valid address */
			if (dst_addr.bcast) goto bcast;

			unicast:;
			for(c = gdl_first(&chan->clients); c != NULL; c = gdl_next(&chan->clients, c)) {
				if (c->id == dst_addr.id) {
					res = insert_msg_in_rlist(c, msg, nonblock, unicast_can_block, 0);
					break;
				}
			}
			break;

		case 'S':
			if (from->master) { /* master->slave */
				if (!dst_addr.valid) return -EFAULT; /* must have a valid address */
				if (dst_addr.bcast) goto bcast;
				unicast_can_block = nonblock; /* nonblocking write should be able to generate EAGAIN */
				goto unicast;
			}
			if (chan->master == NULL) return -ESRCH; /* can't write if there's no master */
			res = insert_msg_in_rlist(chan->master, msg, nonblock, 1, 0); /* slave->master */
			break;

		case 'U':
			if (from->master) goto bcast; /* master->slave */
			if (chan->master == NULL) return -ESRCH; /* can't write if there's no master */
			res = insert_msg_in_rlist(chan->master, msg, nonblock, 1, 0); /* slave->master */
			break;

		case 'B':
			bcast:;

			/* plain broadcast */
			trace(from, "wr dispatch clients");
			for(c = gdl_first(&chan->clients); c != NULL; c = gdl_next(&chan->clients, c)) {
				trace(from, " to:%ld rd=%d", (long)c->id, c->rd);
				if ((c != from) && c->rd)
					res = insert_msg_in_rlist(c, msg, 0, 0, 0);
			}
			break;
	}

	trace(from, " final msg refco: %d", (int)msg->refco);
	return res;
}

static int write_needs_address(channel_t *chan, client_t *from)
{
	switch(chan->type) {
		case 'U': return 0;
		case 'B': return 0;
		case 'A': return 1;
		case 'S': return from->master;
		default:
			/* shouldn't happen */
			logprintf("***ERROR*** wrong channel type %c\n", chan->type);
			return 0;
	}
}

/* A new full message is ready to be dispatched from a client. body is not
   \0 terminated and includes the termiating \n. Return -1 on fatal error
   (e.g. missing address prefix where it is mandatory. */
static int write_dispatch_message(channel_t *chan, client_t *from, char *body, int len, int nonblock)
{
	message_t *msg;
	address_t dst_addr = {0};
	int readers;
	int res;

	if (len <= 1)
		return 0; /* ignore empty messages (multiple \n's) */

	if (write_needs_address(chan, from)) {
		body = addr_parse(&dst_addr, body, &len);
		if (body == NULL)
			return -EFAULT; /* invalid address */
	}

	/* figure if there's anyone to read the message (if not, drop it) */
	readers = write_dispatch_chan_has_readers(chan, from, dst_addr);
	if (readers != 1)
		return readers; /* drop message if nobody would read it or report error */

	msg = message_alloc(chan, body, len);
	if (msg == NULL)
		return -ENOMEM;

	msg->addr = from->id;
	msg->refco++; /* delivery may be succesful for all clients and could free() the msg and we couldn't check afterward */

	res = write_dispatch_chan_msg(chan, from, msg, dst_addr, nonblock);

	/* if we didn't use the message anywhere, or it got delivered, free it.
	   This is the pair of msg->refco++ above */
	message_unref(msg);

	return res;
}

/* Called after new incoming messages has been written into client's wbuf;
   from is the index of where the first search is started for the first \n.
   The buffer may end up with a partial emssage (no terminating \n) and the
   next call after the next write() from the client will be done with a from
   value pointing at the end of that partial message. */
static int write_process(channel_t *chan, client_t *client, long from, int nonblock)
{
	char *sep, *start = client->wbuf.array + from;
	int err;

	while((sep = strchr(start, '\n')) != NULL) {
		if ((err=write_dispatch_message(chan, client, start, sep-start+1, nonblock)) != 0) {
			gds_uninit(&client->wbuf);
			return err;
		}
		*sep = '\0';
		start = sep+1;
	}
	gds_remove(&client->wbuf, 0, start - client->wbuf.array);
	return 0;
}

/* Called from FUSE when client called write() so that new incoming data is
   available. Should return the number of bytes written or -1 on error */
int dispatch_client_write(uint32_t chanID, uint32_t clientID, const char *buf, size_t size, int nonblock)
{
	channel_t *chan;
	client_t *client;
	long from, rem, res;

	chan = id2chan(chanID);
	client = id2client(chan, clientID);
	if (client == NULL)
		return -ENODEV;

	if (size > write_size_max)
		size = write_size_max;

	rem = wbuf_size_max - client->wbuf.used;
	if (rem <= 0)
		return 0;
	if (size > rem)
		size = rem;

	from = client->wbuf.used;
	gds_append_len(&client->wbuf, buf, size);

	if ((res = write_process(chan, client, from, nonblock)) <= 0)
		return res;

	return size;
}

static int read_needs_address(client_t *client)
{
	/* happens to be the same rule as for handlig write() requests */
	return write_needs_address(client->chan, client);
}

/* Render msg into the client's rbuf and unref msg so it can be free'd.
   Return 0 on success. */
static int rbuf_render_unref_message(client_t *client, message_t *msg)
{
	if ((msg->type != MT_CTRL) && read_needs_address(client))
		addr_append(&client->rbuf, msg->addr);
	gds_append_len(&client->rbuf, msg->msg, msg->len);
	message_unref(msg);
	return 0;
}

/* pop first message from the rlist and copy it into the rbuf; returns 0 on
   success, -1 if there's no message available */
static int rlist_pop_and_render(client_t *client)
{
	message_t *msg = NULL;

	mutex_lock(&client->rlist_lock);
	if (client->rlist.used > 0) {
		msg = client->rlist.array[0];
		vtp0_remove(&client->rlist, 0, 1);
	}
	mutex_unlock(&client->rlist_lock);

	if (msg == NULL)
		return -1; /* not an error, don't use -ESOMETHING */

	return rbuf_render_unref_message(client, msg);
}

/* serve a client read() from a non-empty rbuf */
static int rbuf_read(client_t *client, char *buf, size_t size)
{
	if (size >= client->rbuf.used) { /* full message read */
		size = client->rbuf.used;
		memcpy(buf, client->rbuf.array, size);
		client->rbuf.used = 0;
	}
	else { /* partial message read */
		memcpy(buf, client->rbuf.array, size);
		gds_remove(&client->rbuf, 0, size);
	}
	return size;
}

int dispatch_client_read(uint32_t chanID, uint32_t clientID, char *buf, size_t size, int nonblock)
{
	channel_t *chan;
	client_t *client;
	int subseq = 0;
	struct timespec timeout;

	chan = id2chan(chanID);
	client = id2client(chan, clientID);
	if (client == NULL)
		return -ENODEV;

	trace(client, "read %ld", (long)size);

	/* if there's a partial message in rbuf, return that */
	if (client->rbuf.used > 0) {
		trace(client, " from rbuf of %ld", (long)client->rbuf.used);
		return rbuf_read(client, buf, size);
	}

	if (client->rbuf_eof) {
		trace(client, " forced rbuf_eof");
		return 0;
	}

	for(;;) {
		/* if there's anything available in rlist, copy that to rbuf and deliver */
		if (rlist_pop_and_render(client) == 0) {
			trace(client, " from rbuf of %ld after popping message", (long)client->rbuf.used);
			return rbuf_read(client, buf, size);
		}

		/* a zombie channel won't deliver any new message, send EOF */
		if (chan->zombie) {
			trace(client, " ZOMBIE chan -> EOF\n");
			return 0;
		}

		/* nothing to read at the moment, return if non-blocking */
		if (nonblock)
			return -EAGAIN;

		/* nothing in rbuf or rlist; wait for rlist to grow and try again */
		if (!subseq) { /* print only once */
			trace(client, " notify wait");
			subseq = 1;
		}

		clock_gettime(CLOCK_REALTIME, &timeout);
		timeout.tv_nsec += 100000000;
		if (timeout.tv_nsec > 1000000000) {
			timeout.tv_nsec -= 1000000000;
			timeout.tv_sec++;
		}
		notify_wait_timed(&client->rlist_notify, &timeout);
		if (fuse_interrupted())
			return -EINTR;
	}

	return -EIO;
}

int dispatch_client_release(uint32_t chanID, uint32_t clientID)
{
	channel_t *chan;
	client_t *client;

	chan = id2chan(chanID);
	client = id2client(chan, clientID);
	if (client == NULL)
		return -ENODEV;

	client_free(client);
	return 0;
}

static void free_all_chans(gdl_list_t *lst)
{
	channel_t *ch, *next;
	for(ch = gdl_first(lst); ch != NULL; ch = next) {
		next = gdl_next(lst, ch);
		chan_destroy(ch);
	}
}

void dispatch_uninit(void)
{
	free_all_chans(&channels);
	free_all_chans(&zombie_channels);
}
